#!/usr/bin/python
import rospy
from std_msgs.msg import Float64
from custom_msgs.msg import NorthEastHeading, ImuSensor
from geometry_msgs.msg import Vector3, Twist
import numpy as np
from numpy import pi
from numpy import sin
from numpy import cos
import math

from dynamic_reconfigure.server import Server
from dp_controller.cfg import dppidConfig
from custom_msgs.srv import ChangeToDP

# The DP controller calculates the wanted forces and moments based on the reference filter output and estimated states. 
# Inputs: Reference filter values, observer states
# Outputs: Tau vector for thrust allocation

class DP_CONTROLLER(object):

    def __Initialize__(self):
	    # Initialize Publishers
        self.pub_nu_controller = rospy.Publisher('nu_controller', NorthEastHeading, queue_size=10)
        self.pub_tau_controller = rospy.Publisher('tau_controller', NorthEastHeading, queue_size=10)
        self.PID_NORTH_pub = rospy.Publisher('tau_Error_North', NorthEastHeading, queue_size=10)
        self.PID_EAST_pub = rospy.Publisher('tau_Error_East', NorthEastHeading, queue_size=10)
        self.PID_PSI_pub = rospy.Publisher('tau_Error_Psi', NorthEastHeading, queue_size=10)

        # Initialize Subsribers
        rospy.Subscriber("eta_d", NorthEastHeading, self.eta_d_callback)
        rospy.Subscriber("eta_GNSS", NorthEastHeading, self.eta_obs_callback)  # [m,m,deg]
        rospy.Subscriber('observer/nu', Twist, self.nu_obs_callback)
        rospy.Subscriber("observer/orientation", Vector3 , self.attitude_obs_callback)
        rospy.Subscriber("xsens/imu", ImuSensor, self.nu_sim_callback)
        DPss = rospy.Service("change_to_DP_pid", ChangeToDP, self.dp_reset_callback)        
        srv = Server(dppidConfig, self.parameter_callback)
		
		# Initialize matrices and parameters
		self.eta_d = np.array([0, 0, 0])
        self.eta_d = self.eta_d.reshape(3, 1)
        self.dot_eta_d = np.array([0, 0, 0])
        self.dot_eta_d = self.dot_eta_d.reshape(3, 1)
        self.ddot_eta_d = np.array([0, 0, 0])
        self.ddot_eta_d = self.ddot_eta_d.reshape(3, 1)
        self.eta_obs = np.array([0, 0, 0])
        self.eta_obs = self.eta_obs.reshape(3, 1)
        self.heading = 0
        self.nu_obs = np.array([0, 0, 0]) 
        self.nu_obs = self.nu_obs.reshape(3, 1)
        self.eta_previous = np.array([0, 0, 0])
        self.eta_previous = self.eta_previous.reshape(3, 1)
        self.eta_int = np.array([0, 0, 0])
        self.eta_int = self.eta_int.reshape(3, 1)
		
		
        self.warmed_up = 0

		
        self.nu_sim = np.array([0, 0, 0])
        self.nu_sim = self.nu_sim.reshape(3, 1)



        self.h = 0  # Time Elapsed since last iteration

		# Max values for tau vector
        self.tau_sat = np.array([69, 45, 75])
        self.tau_sat = self.tau_sat.reshape(3, 1)

		
		# Gain values initializaation (These WILL be overwritten by the dynamic reconfigure panel upon start) 
        Kp_matrix = np.array([1, 1, 1])
        self.Kp = np.diag(Kp_matrix)
        Kd_matrix = np.array([1, 1, 1])
        self.Kd = np.diag(Kd_matrix)
        Ki_matrix = np.array([1, 1, 1])
        self.Ki = np.diag(Ki_matrix)
        self.windup_var = np.array([0, 0, 0])
        self.windup_var = self.windup_var.reshape(3, 1)

        # Initialize ROS Node
        rospy.Initialize_node('DPcontrollerNode', anonymous=True)

        self.time_prev = rospy.get_time()

		
	# This callback function is ran everytime a new reference filter message is recieved, and is the core of the controller. 
    def eta_d_callback(self, eta_d_input):
        DEG2RAD = np.pi / 180
        RAD2DEG = 180 / np.pi
        self.h = (rospy.get_time() - self.time_prev)
        self.time_prev = rospy.get_time()

        if self.warmed_up == 1:
            self.eta_d = np.array([eta_d_input.pos_north, eta_d_input.pos_east, eta_d_input.pos_heading * DEG2RAD])
            self.eta_d = self.eta_d.reshape(3, 1)
            # Wrap yaw within +/- pi(180 deg)
            self.eta_d[2, 0] = np.remainder(self.eta_d[2, 0], 2 * pi)
            if self.eta_d[2, 0] > pi:
                self.eta_d[2, 0] = self.eta_d[2, 0] - 2 * pi
            elif self.eta_d[2, 0] < -pi:
                self.eta_d[2, 0] = self.eta_d[2, 0] + 2 * pi
            
            try: 
                self.eta_obs[2,0] = self.heading
            except IndexError:
                print(self.eta_obs)
                print(self.heading)
				
            # Rotation Matrix
            R11 = np.cos(self.eta_obs[2, 0])
            R12 = -np.sin(self.eta_obs[2, 0])
            R13 = 0
            R21 = np.sin(self.eta_obs[2, 0])
            R22 = np.cos(self.eta_obs[2, 0])
            R23 = 0
            R31 = 0
            R32 = 0
            R33 = 1
            R = np.array([R11, R12, R13, R21, R22, R23, R31, R32, R33])
            R = R.reshape(3, 3)
            R_trans = np.transpose(R)

            obs_eta_body = R_trans.dot(self.eta_obs)
            eta_body_d = R_trans.dot(self.eta_d)
            self.dot_eta_d = np.array([eta_d_input.vel_north, eta_d_input.vel_east, eta_d_input.vel_heading * DEG2RAD])
            self.dot_eta_d = self.dot_eta_d.reshape(3, 1)
            nu_d = R_trans.dot(self.dot_eta_d)
			# Skew Symmetric Matrix 
            S11 = 0
            S12 = -nu_d[2]
            S13 = 0
            S21 = nu_d[2]
            S22 = 0
            S23 = 0
            S31 = 0
            S32 = 0
            S33 = 0
            S = np.array([S11, S12, S13, S21, S22, S23, S31, S32, S33])
            skewmat = S.reshape(3, 3)
			
			
			
            self.ddot_eta_d = np.array([eta_d_input.acc_north, \
                eta_d_input.acc_east, eta_d_input.acc_heading * DEG2RAD])
            self.ddot_eta_d = self.ddot_eta_d.reshape(3, 1)
            nu_dot_d = R_trans.dot(self.ddot_eta_d) + skewmat.dot(R_trans).dot(self.dot_eta_d)

            # Error Matrix
            eta_error = obs_eta_body - eta_body_d

            # Wrap yaw error within +-pi (180 degrees)
            eta_error[2, 0] = np.remainder(eta_error[2, 0], 2 * pi)
            if eta_error[2, 0] > pi:
                eta_error[2, 0] = eta_error[2, 0] - 2 * pi
            elif eta_error[2, 0] < -pi:
                eta_error[2, 0] = eta_error[2, 0] + 2 * pi

				
            # Compute Tau
            tau_PID = -(self.Kp.dot(eta_error) + self.Ki.dot(self.eta_int) + self.Kd.dot(self.nu_obs - nu_d))
            
            PID_PSI = NorthEastHeading()
            PID_PSI.pos_north = (eta_error)[2,0]
            PID_PSI.pos_east = (self.eta_int)[2,0]
            PID_PSI.pos_heading = (self.nu_obs - nu_d)[2,0]
            self.PID_PSI_pub.publish(PID_PSI)
            
            PID_EAST = NorthEastHeading()
            PID_EAST.pos_north = (eta_error)[1,0]
            PID_EAST.pos_east = (self.eta_int)[1,0]
            PID_EAST.pos_heading = (self.nu_obs - nu_d)[1,0]
            self.PID_EAST_pub.publish(PID_EAST)
            
            PID_NORTH = NorthEastHeading()
            PID_NORTH.pos_north = (eta_error)[0,0]
            PID_NORTH.pos_east = (self.eta_int)[0,0]
            PID_NORTH.pos_heading = (self.nu_obs - nu_d)[0,0]
            self.PID_NORTH_pub.publish(PID_NORTH)
            tau = tau_PID



            # Output saturation
            if np.absolute(tau[0, 0]) > self.tau_sat[0, 0] or np.absolute(tau[1, 0]) > self.tau_sat[1, 0] or np.absolute(tau[2, 0]) > self.tau_sat[2, 0]:
                if np.absolute(tau[0, 0]) > self.tau_sat[0, 0]:
                    tau[2] = np.sign(tau[2, 0]) * np.absolute(self.tau_sat[0, 0] / tau[0, 0]) * np.absolute(tau[2, 0])
                    tau[1] = np.sign(tau[1, 0]) * np.absolute(self.tau_sat[0, 0] / tau[0, 0]) * np.absolute(tau[1, 0])
                    tau[0] = np.sign(tau[0, 0]) * self.tau_sat[0, 0]
                if np.absolute(tau[1, 0]) > self.tau_sat[1, 0]:
                    tau[2] = np.sign(tau[2, 0]) * np.absolute(self.tau_sat[1, 0] / tau[1, 0]) * np.absolute(tau[2, 0])
                    tau[0] = np.sign(tau[0, 0]) * np.absolute(self.tau_sat[1, 0] / tau[1, 0]) * np.absolute(tau[0, 0])
                    tau[1] = np.sign(tau[1, 0]) * self.tau_sat[1]
                if np.absolute(tau[2, 0]) > self.tau_sat[2, 0]:
                    tau[1, 0] = np.sign(tau[1, 0]) * np.absolute(self.tau_sat[2, 0] / tau[2, 0]) * np.absolute(tau[1, 0])
                    tau[0, 0] = np.sign(tau[0, 0]) * np.absolute(self.tau_sat[2, 0] / tau[2, 0]) * np.absolute(tau[0, 0])
                    tau[2, 0] = np.sign(tau[2, 0]) * self.tau_sat[2, 0]

            # Anti-Windup (Integrals)

            self.eta_int = self.eta_int + self.h * (eta_error + self.windup_var)

            self.windup_var = np.array([0, 0, 0])
            self.windup_var = self.windup_var.reshape(3, 1)
			
            if self.eta_int[0, 0] > self.tau_sat[0, 0]:
                self.windup_var[0, 0] = self.tau_sat[0, 0] - self.eta_int[0, 0]
            elif self.eta_int[0, 0] < -self.tau_sat[0, 0]:
                self.windup_var[0, 0] = -self.tau_sat[0, 0] - self.eta_int[0, 0]

            if self.eta_int[1, 0] > self.tau_sat[1, 0]:
                self.windup_var[1, 0] = self.tau_sat[1, 0] - self.eta_int[1, 0]
            elif self.eta_int[1, 0] < -self.tau_sat[1, 0]:
                self.windup_var[1, 0] = -self.tau_sat[1, 0] - self.eta_int[1, 0]

            if self.eta_int[2, 0] > self.tau_sat[2, 0]:
                self.windup_var[2, 0] = self.tau_sat[2, 0] - self.eta_int[2, 0]
            elif self.eta_int[2, 0] < -self.tau_sat[2, 0]:
                self.windup_var[2, 0] = -self.tau_sat[2, 0] - self.eta_int[2, 0]

            self.eta_previous = self.eta_obs

            nu_controller = NorthEastHeading()
            nu_controller.pos_north = self.nu_obs[0]
            nu_controller.pos_east = self.nu_obs[1]
            nu_controller.pos_heading = self.nu_obs[2]
            self.pub_nu_controller.publish(nu_controller)

			
            # Publish tau_controller
            tau_controller = NorthEastHeading()
            tau_controller.pos_north = tau[0, 0]
            tau_controller.pos_east = tau[1, 0]
            tau_controller.pos_heading = tau[2, 0]
            self.pub_tau_controller.publish(tau_controller)

    def nu_obs_callback(self, nu_obs_input):
        DEG2RAD = np.pi / 180
        RAD2DEG = 180 / np.pi

        self.nu_obs = np.array([nu_obs_input.linear.x,
                                nu_obs_input.linear.y,
                                nu_obs_input.angular.z])
        self.nu_obs = self.nu_obs.reshape(3, 1)

    def nu_sim_callback(self, nu_simBody_input):
        DEG2RAD = np.pi / 180
        RAD2DEG = 180 / np.pi
        self.nu_sim = np.array([nu_simBody_input.linear_velocity.x,
                                nu_simBody_input.linear_velocity.y,
                                nu_simBody_input.angular_velocity.z])
        self.nu_sim = self.nu_sim.reshape(3, 1)	
    
    def attitude_obs_callback(self, attitude_obs_input):
        self.heading = attitude_obs_input.z*(np.pi / 180)
        
        self.heading = np.remainder(self.heading, 2 * np.pi)
        if self.heading > np.pi:
            self.heading = self.heading - 2*np.pi
        elif self.heading < -np.pi:
            self.heading = self.heading + 2*np.pi
    def dp_reset_callback(self, data):
        elapsed_time = 0
        start_time = rospy.get_time() 
        while elapsed_time < 10:
            elapsed_time = rospy.get_time()-start_time
        self.warmed_up = 1
		
    def eta_obs_callback(self, eta_intnput):
        DEG2RAD = np.pi / 180
        RAD2DEG = 180 / np.pi

        self.eta_obs = np.array([eta_intnput.pos_north, eta_intnput.pos_east, eta_intnput.pos_heading * DEG2RAD])
        self.eta_obs = self.eta_obs.reshape(3, 1)

        # Wrap yaw measured within +/- pi(180 deg)
        self.eta_obs[2, 0] = np.remainder(self.eta_obs[2, 0], 2 * pi)
        if self.eta_obs[2, 0] > pi:
            self.eta_obs[2, 0] = self.eta_obs[2, 0] - 2 * pi
        elif self.eta_obs[2, 0] < -pi:
            self.eta_obs[2, 0] = self.eta_obs[2, 0] + 2 * pi

    def parameter_callback(self, config, level):
		# This callback function takes input from the dynamic reconfigure control panel,
		# and changes the gain values accordingly.

        # Update parameters
        Kp_matrix = np.array([config.Kp_surge, config.Kp_sway, config.Kp_heading])
        self.Kp = np.diag(Kp_matrix)
        Kd_matrix = np.array([config.Kd_surge, config.Kd_sway, config.Kd_heading])
        self.Kd = np.diag(Kd_matrix)
        Ki_matrix = np.array([config.Ki_surge, config.Ki_sway, config.Ki_heading])
        self.Ki = np.diag(Ki_matrix)


        rospy.loginfo('Config setup complete! ')

        return config


if __name__ == '__main__':
    try:
        # Construct DP_CONTROLLER object
        DP_CONTROLLER()
        # Let ROS take over
        rospy.spin()
    except rospy.ROSInterruptException:
        pass
