#!/usr/bin/python
from __future__ import division
import rospy
from std_msgs.msg import *  # Imports all std_msgs
from custom_msgs.msg import NorthEastHeading
from custom_msgs.msg import podAngle
from custom_msgs.msg import SternThrusterSetpoints
from custom_msgs.msg import bowControl
from custom_msgs.msg import diffThrottleStern
from custom_msgs.msg import ThrusterFaults
from numpy import *
import signal
from numpy.matlib import sin
from numpy.matlib import cos
from numpy.matlib import hstack
from numpy.matlib import vstack
from numpy.matlib import sqrt
from numpy.matlib import isnan
from numpy.linalg import pinv
import py4j.java_gateway as java_gateway

from py4j.java_gateway import JavaGateway, GatewayParameters
import time


class Thruster_Allocation(object):

    def __init__(self):
        java_gateway.launch_gateway(create_new_process_group=False)
        self.gateway = JavaGateway()
        
        # Init process ID handle
        self.pro = 'NaN'
        # Init FTC change variable
        self.newFTCmsg = 0
        # Init ROS Node
        rospy.init_node('thrusterAllocNode', anonymous=True)
        
        # Init thruster paramaters 
        self.SternStarFault = 0
        self.SternPortFault = 0
        self.BowFault = 0
        # Init Publishers
        self.pub_stern_angles = rospy.Publisher('pod_angle_input', podAngle, queue_size=1)
        self.pub_stern_thruster_setpoints = rospy.Publisher("stern_thruster_setpoints", SternThrusterSetpoints, queue_size=1)
        self.pub_bow_control = rospy.Publisher("bow_control", bowControl, queue_size=1)

        # Init Subscribers
        # Subscriber for control forces from either DP controller or RC remote (manual T.A.)
        rospy.Subscriber("tau_controller", NorthEastHeading, self.tau_controller_callback)
        # Subscriber for differential throttle used in manual mode
        rospy.Subscriber("diff_throttle_stern", diffThrottleStern, self.diff_throttle_stern_callback)
        # Subscriber for control effort from heading controller, controlling stern steppers
        rospy.Subscriber("heading_controller_output", Float64, self.heading_control_effort_callback)
        # Subscriber for control effort from speed controller
        rospy.Subscriber("speed_controller_output", Float64, self.speed_controller_effort_callback)
        # Subscriber for thruster fault tolerant control
        rospy.Subscriber("actuator_faults_predicted", ThrusterFaults, self.thruster_FTC_callback)

    def norm(self, val_input, val_min, val_max, out_min, out_max):
        output = int((val_input - val_min) * (out_max - out_min) / (val_max - val_min) + out_min)
        return output

    def sanity_check(self, input_check):
        if input_check < -100:
            return -100
        elif input_check > 100:
            return 100
        else:
            return input_check

    def solv(self, f, a, b, c):
        c = c - f
        if b ** 2 - 4 * a * c >= 0:
            if abs(sqrt(b ** 2 - 4 * a * c)) >= b:
                return ((-b + sqrt(b ** 2 - 4 * a * c)) / (2 * a))
        else:
            rospy.loginfo('cannot solve for demanded thrust')

    ''' Callback for control effort from heading controller, used in Heading Control Mode '''
    def heading_control_effort_callback(self, control_effort_in):
        control_effort_out = control_effort_in.data
        pod_angle = podAngle()  # Declare pod_angle_input as a class
        control_effort_angle = -control_effort_out  # Negative sign here to turn the pods the right way.

        # Saturation of control output
        rospy.logwarn("HeadingControl Callback in ThrustAlloc")
        saturation = 45
        if control_effort_angle > saturation:
            control_effort_angle = saturation
        if control_effort_angle < -saturation:
            control_effort_angle = -saturation

        # ~ rospy.loginfo(control_effort_angle)

        # Assign control effort to port and starboard steppers
        pod_angle.port = float(control_effort_angle)
        pod_angle.star = float(control_effort_angle)

            # Publish pod_angle
        self.pub_stern_angles.publish(pod_angle)

    ''' Callback for topic /speed_controller_output '''
    def speed_controller_effort_callback(self, control_input):
        # Create message to publish
        stern_thruster_setpoints = SternThrusterSetpoints()

        # Perform sanity check of received value
        throttle_in = self.sanity_check(control_input.data)

        # Add value to message
        stern_thruster_setpoints.star_effort = throttle_in
        stern_thruster_setpoints.port_effort = throttle_in

        # Publish Message to topic: /stern_thruster_setpoints
        self.pub_stern_thruster_setpoints.publish(stern_thruster_setpoints)

    ''' Callback for manual control mode '''
    def diff_throttle_stern_callback(self, diff_throttle_input):
        # Performs sanity check
        stern_thruster_setpoints = SternThrusterSetpoints()
        throttle_in = self.sanity_check(diff_throttle_input.throttle) # Not self
        rudder_in = self.sanity_check(diff_throttle_input.rudder)

        # Mixing of throttle and rudder inputs
        stern_thruster_setpoints.port_effort = float(throttle_in) + float(rudder_in)
        stern_thruster_setpoints.star_effort = float(throttle_in) - float(rudder_in)
        # Note: Port propeller is mirrored physically and needs to rotate the other way

        # Publish Throttle to topics
        self.pub_stern_thruster_setpoints.publish(stern_thruster_setpoints)


    def thruster_FTC_callback(self, thruster_info):
		self.prevSFval = self.SternStarFault
		self.prevPFval = self.SternPortFault
		self.prevBFval = self.BowFault
		self.SternStarFault = thruster_info.star_frozen
		self.SternPortFault = thruster_info.port_frozen
		self.BowFault = thruster_info.bow_frozen
		
		if((self.prevSFval == 0 and self.SternStarFault == 1) or (self.prevPFval == 0 and self.SternPortFault == 1) or (self.prevBFval == 0 and self.BowFault == 1)):
			import os
			import subprocess
			# Kill process ID of the JVM spawning thrustAlloc instances
			os.killpg(os.getpgid(self.pro.pid), signal.SIGTERM)
			
			# Signal thrust allocation change
			self.newFTCmsg = 1

    ''' Callback for DP and Manual Thrust allocation mode '''
    def tau_controller_callback(self, tau_d):
            
        # Thruster setup
        #   _________________________________
        #   |                                 \
        #   |   X-  (a1)                        \
        #   |                       X-  (a3)     )
        #   |   X-  (a2)                        /
        #   |________________________________ /

            # ~ rospy.loginfo('tau_d.north: %d , tau_d.east: %d,  tau_d.heading: %d ,',
            # tau_d.north, tau_d.east, tau_d.heading);
            # Try to run thrusteralloc, except, start thrusteralloc and run.
        
        # Give the java subprocess some time to shut down upon change in thruster alloc
        if(self.newFTCmsg == 1):
            time.sleep(2.0)
            self.newFTCmsg = 0
        
        # Find the amount of faulty thrusters
        faultCount = self.SternStarFault + self.SternPortFault + self.BowFault 
        if(faultCount > 0):
			print("Fault has occured!")
        # If there is a fault in the thrust arrangement, change to a different thrust allocation in the bank of TA algorithms
        
        if(faultCount == 1):
            if(self.SternStarFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarDisabled.java already started')
                    rospy.logwarn("Stern Starboard Thruster Disengaged!")
                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_StarDisabled', shell=True, stdout=subprocess.PIPE, preexec_fn=os.setsid) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)
            elif(self.SternPortFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_PortDisabled.java already started')
                    rospy.logwarn("Stern Port Thruster Disengaged!")
                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_PortDisabled', shell=True, stdout=subprocess.PIPE) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_PortDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)
            elif(self.BowFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('ThrusterAllocation_BowDisabled.java already started')
                    rospy.logwarn("Bow Thruster Disengaged!")

                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_BowDisabled', shell=True, stdout=subprocess.PIPE) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_BowDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)
        # In case of multiple faults
        elif(faultCount == 2):
            if(self.SternStarFault == 1 and self.SternPortFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarAndPortDisabled.java already started')
                    rospy.logwarn("Stern Port and Starboard Thruster Disengaged!")
                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_StarAndPortDisabled', shell=True, stdout=subprocess.PIPE) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarAndPortDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)
            elif(self.SternStarFault == 1 and self.BowFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarAndBowDisabled.java already started')
                    rospy.logwarn("Stern Star and Bow Thruster Disengaged!")
                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_StarAndBowDisabled', shell=True, stdout=subprocess.PIPE) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_StarAndBowDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)
            elif(self.SternPortFault == 1 and self.BowFault == 1):
                try:
                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_PortAndBowDisabled.java already started')
                    rospy.logwarn("Stern Port and Bow Thruster Disengaged!")
                except:
                    # Move working directory to thrusteralloc folder
                    import os
                    wd = os.getcwd()
                    os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                    import subprocess
                    process = subprocess.Popen('java -Djava.ext.dirs=. TA_PortAndBowDisabled', shell=True, stdout=subprocess.PIPE) #run java process
                    self.pro = process
                    # Give the Java Gateway time to start
                    time.sleep(3.0)

                    u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                    rospy.loginfo('TA_PortAndBowDisabled.java started')

                    # Working directory back to original
                    os.chdir(wd)                
        
        # In case of no fault: 
        else:
            try:
                u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                rospy.loginfo('ThrusterAllocation.java already started')
            except:
                # Move working directory to thrusteralloc folder
                import os
                wd = os.getcwd()
                os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/dp_controller/nodes/ThrusterAlloc')
                java_gateway.launch_gateway()
                self.gateway = JavaGateway()
                import subprocess
                process = subprocess.Popen('java -Djava.ext.dirs=. ThrusterAllocation', shell=True, stdout=subprocess.PIPE, preexec_fn=os.setsid) #run java process
                self.pro = process
                # Give the Java Gateway time to start
                time.sleep(3.0)
                u = self.gateway.entry_point.runTA(float(tau_d.pos_north), float(tau_d.pos_east), float(tau_d.pos_heading))
                rospy.loginfo('ThrusterAllocation.java started')
                # Working directory back to original
                os.chdir(wd)
                
        # log desired force for each propeller in newton and angles
        rospy.loginfo('u.F1: %s , u.F2: %s,  u.F3: %s , u.a1: %s , u.a2: %s,  u.a3: %s ,',
                      u[0], u[1], u[2], u[3], u[4], u[5])

        # Constant K values F = K*n*n
        if u[2] >= 0:
            K = array([0.0027, 0.0027, 0.001518])
        else:
            K = array([0.0027, 0.0027, 0.0006172])

        # ~ #Calculate n [rpm] : f = K*n*abs(n)
        Tf = array([u[0], u[1], u[2]])
        n = sign(Tf / K) * sqrt(abs(Tf / K))
        # ~ rospy.loginfo(n)
        rospy.loginfo('n.Port: %s , n.Star: %s, n.Bow: %s  ', n[0], n[1], n[2])

            # Init Custom Messages
        pod_angle = podAngle()
        stern_thruster_setpoints = SternThrusterSetpoints()
        bow_control = bowControl()

        # Fill Pod Angle Custom Message
        pod_angle.port = float(u[3] * 180 / pi)
        pod_angle.star = float(u[4] * 180 / pi)

        # Fill Throttle Custom Message
        stern_thruster_setpoints.star_effort = float(n[1])
        stern_thruster_setpoints.port_effort = float(n[0])
        bow_spd = float(n[2])
        # Bow throttle doesn't work at low inputs. (DC Motor)
        if bow_spd > 0:
            bow_spd = float(bow_spd)
        if bow_spd < 3 and bow_spd > -3:
            bow_spd = float(bow_spd)
        else:
            bow_spd = float(bow_spd)

        # Linear Actuator For Bow Pod
        act_out = 2  # 2 = Go down
        #rospy.logwarn("u[5] %f",u[5] * 180 / pi)
        
        ## This part is specific for simulator use - may need to be reverted for real life testing!
        servo_out = self.norm(u[5] * 180 / pi, -180, 180, -180, 180) # This mapping was broken for simulator use.
        #rospy.logwarn("servo_out %f",servo_out)
        # [-100 100]
        #if servo_out < -100:
        #    servo_out = -100
        #if servo_out > 100:
        #    servo_out = 100

        ## Fill bow control custom message
        bow_control.throttle_bow = bow_spd
        bow_control.position_bow = servo_out
        bow_control.lin_act_bow = act_out

        # ~ rospy.loginfo('angle.Bow: %s , angle.Port: %s , angle.Star: %s ', a1, a2, a3);
        self.pub_stern_angles.publish(pod_angle)
        self.pub_stern_thruster_setpoints.publish(stern_thruster_setpoints)
        self.pub_bow_control.publish(bow_control)
    def shutdown_imminent(self):
        import os
        os.killpg(os.getpgid(self.pro.pid), signal.SIGTERM)


if __name__ == '__main__':
    try:
        # Init Thruster_Allocation
        Allocator = Thruster_Allocation()
        rospy.on_shutdown(Allocator.shutdown_imminent)
        # Let ROS take over
        rospy.spin()
    except rospy.ROSInterruptException:
        #rospy.logwarn("Quittan")
        #Allocator.gateway.shutdown()
        #Allocator.gateway.jvm.System.exit(0)
        #import os
        #os.killpg(os.getpgid(Allocator.pro.pid), signal.SIGTERM)
        pass
