#!/usr/bin/python

import numpy as np
import pandas as pd
import rospy
import os
import time
import pickle
from custom_msgs.msg import gnssGGA, inertialSensor, gnssOutlier, imuOutlier
from std_msgs.msg import Header
from sklearn.svm import OneClassSVM

class ocsvm_model_gnss(object):

    def __init__(self, nu=0.001, gamma=5.9, retrainThreshold = 4500):
        self.nu = nu
        self.gamma = gamma
        self.retrainThreshold = retrainThreshold
        self.storageDF = pd.DataFrame(columns=['Timestamp','Latitude', 'Longitude', 'Altitude'])
        
        gnssSub = rospy.Subscriber("/signal/GNSS_check_outlier", gnssGGA, self.gnssCB, queue_size=1, buff_size = 125536) #Increased the buffer size to prevent delays        
        self.GNSSoutlierPub = rospy.Publisher("/signal/GNSS_outlier_report", gnssOutlier, queue_size=1)
        
        rospy.logwarn("OCSVM inited")
    def model_initialize(self, D_initial_training):        

        oneclass_svm = OneClassSVM(nu=self.nu, gamma=self.gamma)
        oneclass_svm.fit(D_initial_training)
        self.model_ = oneclass_svm
        
        rospy.logwarn("OCSVM Model initialized")
    def model_reload(self, model):
        self.model_ = model
        print("Previous GNSS outlier detection model detected")
    def model_update(self):
         # Pick out the three variables we are interested in
         newDatapoints = self.storageDF.iloc[:,[1,2,3]]
         
         support_vectors_current = self.model_.support_vectors_.copy()
         newTrainingset = np.vstack((support_vectors_current, newDatapoints))   
         self.model_ = self.model_.fit(newTrainingset)
         
         # Empty and create a new storageDF container
         del self.storageDF
         self.storageDF = pd.DataFrame(columns=['Timestamp','Latitude', 'Longitude', 'Altitude'])

         
    
    def inlierPrediction(self, ts,lat,lon,alt):
        alt = 0.0
        newDatapoint = pd.DataFrame([[ts,lat,lon,alt]],columns=['Timestamp','Latitude', 'Longitude', 'Altitude'])
        inlier = self.model_.predict(newDatapoint.iloc[:,[1,2,3]])[0]
        if(inlier == 1):
            self.storageDF = self.storageDF.append(newDatapoint,ignore_index=True)
            if(self.storageDF.shape[0] >= self.retrainThreshold):
                print("time to update")
                self.model_update()
        
        return inlier;
    
    def gnssCB(self,data):
        
        if(hasattr(self, "model_")):
            #print("This will now enter the model:")
            #print("Lat Lon Alt", data.latitude,data.longitude,data.altitude)
            prediction = self.inlierPrediction(data.header.stamp, data.latitude,data.longitude,data.altitude)
            #print("The model deems it as: ", prediction)
            # Fill message and publish
            outlier_msg = gnssOutlier()
            h = data.timestamp
            outlier_msg.timestamp = h
            outlier_msg.gnss_unit = data.gnss_unit
            outlier_msg.outlier = prediction
            self.GNSSoutlierPub.publish(outlier_msg)

    def shutdown_imminent(self): # Save/dump model when the central ROS node is shut down
        os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/ml_pack')
        filename = "NewModel_GNSS_OCSVM" + str(int(time.time())) + "_.p"
        outfile = open(filename,'wb')
        pickle.dump(self.model_,outfile,protocol=2)
        outfile.close()
        rospy.sleep(3)
        print("GNSS OCSVM Model has been saved!")

class ocsvm_model_imu(object):


    def __init__(self, nu=0.01, gamma=0.9, retrainThreshold = 400):
        self.nu = nu
        self.gamma = gamma
        self.retrainThreshold = retrainThreshold
        self.storageDF = pd.DataFrame(columns=['Timestamp','Roll', 'Pitch', 'Yaw','rollRate','pitchRate','yawRate'])
        
        imuSub = rospy.Subscriber("/signal/IMU_check_outlier", inertialSensor, self.imuCB, queue_size=1, buff_size = 125536)
        
        self.IMUoutlierPub = rospy.Publisher("/signal/IMU_outlier_report", imuOutlier, queue_size=1)
        
        rospy.logwarn("OCSVM for IMU units inited")
    def model_initialize(self, D_initial_training):        

        oneclass_svm = OneClassSVM(nu=self.nu, gamma=self.gamma)
        oneclass_svm.fit(D_initial_training)
        self.model_ = oneclass_svm
        
        rospy.logwarn("OCSVM Model initialized")
    def model_reload(self, model):
        self.model_ = model
        print("Previous IMU outlier detection model detected")
    def model_update(self):
         # Pick out the variables we are interested in
         newDatapoints = self.storageDF.iloc[:,2:]
         
         support_vectors_current = self.model_.support_vectors_.copy()
         newTrainingset = np.vstack((support_vectors_current, newDatapoints))   
         self.model_ = self.model_.fit(newTrainingset)
         
         # Empty and create a new storageDF container
         del self.storageDF
         self.storageDF = pd.DataFrame(columns=['Timestamp','Roll', 'Pitch', 'Yaw','rollRate','pitchRate','yawRate'])

         
    
    def inlierPrediction(self,ts, roll,pitch,yaw,rollRate,pitchRate, yawRate):
        newDatapoint = pd.DataFrame([[ts, roll,pitch,yaw,rollRate,pitchRate,yawRate]],columns=['Timestamp','Roll', 'Pitch', 'Yaw','rollRate','pitchRate','yawRate'])
        inlier = self.model_.predict(newDatapoint.iloc[:,2:])[0]
        
        if(inlier == 1):
            self.storageDF = self.storageDF.append(newDatapoint,ignore_index=True)
            if(self.storageDF.shape[0] >= self.retrainThreshold):
                self.model_update()
        return inlier;

    def imuCB(self,data):
        
        if(hasattr(self, "model_")):
            #print("This will now enter the model:")
            prediction = self.inlierPrediction(data.header.stamp, data.roll,data.pitch,data.yaw,data.rollRate,data.pitchRate, data.yawRate)
            #print("The model deems it as: ", prediction)
            
            # Fill message and publish
            outlier_msg = imuOutlier()
            h = data.timestamp
            outlier_msg.timestamp = h
            outlier_msg.imu_unit = data.imuUnit
            outlier_msg.outlier = prediction
            self.IMUoutlierPub.publish(outlier_msg)

    def shutdown_imminent(self): # Save/dump model when the central ROS node is shut down
        os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/ml_pack')
        filename = "newmodel_IMU_OCSVM" + str(int(time.time())) + "_.p"
        outfile = open(filename,'wb')
        pickle.dump(self.model_,outfile,protocol=2)
        outfile.close()
        print("Model has been saved!")




def main():
    # initiate node
    rospy.init_node('OutlierDetectionNode')
    load_pickle = 1 # Flag to choose whether to reload an old model or train a new one

    # Create One-Class SVM objects
    OCSVM_GNSS = ocsvm_model_gnss()
    OCSVM_IMU = ocsvm_model_imu()

    if(load_pickle == 1): # Reload old machine learning model (no training)
        filepath_gnss = "/home/ubuntudynamo/Desktop/Revolt/src/ml_pack/NewModel_GNSS_OCSVM1559040635_.p" #outlierdetector_model.p"
        infile = open(filepath_gnss)
        model_reloaded = pickle.load(infile)
        infile.close()
        OCSVM_GNSS.model_reload(model_reloaded)
        filepath_imu = "/home/ubuntudynamo/Desktop/Revolt/src/ml_pack/imu_outlierdetector_model.p" #imu_outlierdetector_model.p"
        infile_imu = open(filepath_imu)
        model_reloaded = pickle.load(infile_imu)
        infile_imu.close()
        OCSVM_IMU.model_reload(model_reloaded)
    else: # Train new machine learning model on basis of old data (no previous model) - NB: all data must be inliers! 
        
        # Read CSV file containing initial training data
        os.chdir('/home/ubuntudynamo/Desktop/Revolt/src/ml_pack')

        # Create a dataframe object containing the initial training data
        df_train_gnss = pd.read_csv('gnss_training_data_nofaults.csv')
        df_train_gnss.fieldaltitude.map(lambda x: '{:.0f}'.format(x))

        # Select the lat, lon and alt variables from the training data
        X_train_gnss = df_train_gnss.iloc[:, [5,6,9]]    
        OCSVM_GNSS.model_initialize(X_train_gnss)
        
        # Same procedure as above, but with IMU data 
        df_train_imu = pd.read_csv('imu_trainingdata.csv')
        X_train_imu = df_train_imu.iloc[:, 1:]
        OCSVM_IMU.model_initialize(X_train_imu)




    rospy.on_shutdown(OCSVM_GNSS.shutdown_imminent)
    rospy.on_shutdown(OCSVM_IMU.shutdown_imminent)
    rospy.spin()

    
    
if __name__ == '__main__':
    try:
        main()
    except rospy.ROSInterruptException:
        pass
