#include "ros/ros.h"
#include "std_msgs/String.h"
#include <std_msgs/Float64.h>
#include "std_msgs/Header.h"
#include "std_msgs/UInt8.h"
#include "std_msgs/Time.h"
#include "custom_msgs/gnssGGA.h"
#include "custom_msgs/gpsFIX.h"
#include "custom_msgs/gnssOutlier.h"
#include "custom_msgs/imuOutlier.h"
#include "custom_msgs/residualPose.h"
#include "custom_msgs/residualIMU.h"
#include "custom_msgs/ChangeToDP.h"
#include "custom_msgs/inertialSensor.h"
#include <sstream>
#include <cmath>
#include <numeric>
#include <algorithm>
#include <math.h>
#include <boost/math/distributions/normal.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>

class SigBlock
{
public:
	explicit SigBlock(ros::NodeHandle nh);
    int NUM_GNSS_UNITS = 3;
    int NUM_IMU_UNITS = 3;
    
    struct SIG_INFO {
        bool LostSignal;
        bool BiasedSignal;
        bool FrozenSignal;
        bool HighVarSignal;
        bool DriftingSignal;
        bool OutlierAlarm;
        bool ScalingError;
        bool InvalidRange;
        bool DeviationVoting;
    };
    typedef std::vector<SIG_INFO> SIG_INFO_VECTOR;
    
    struct GNSS_CONTAINER {
        double latitude;
        double longitude;
        double altitude;
        double timestamp_double;
        ros::Time timestamp;
        bool deactivated;
        bool published;
    };
    typedef std::vector<GNSS_CONTAINER> GNSS_DATA_VECTOR;
    
    struct IMU_CONTAINER {
        double roll;
        double pitch;
        double yaw;
        double rollRate;
        double pitchRate;
        double yawRate;
        double timestamp_double;
        ros::Time timestamp;
        bool deactivated;
        bool published;
    };
    typedef std::vector<IMU_CONTAINER> IMU_DATA_VECTOR;

    struct S_vector {
        double measurement;
        double S_high;
        double S_low;
    };
    struct GNSS_RESIDUAL_CONTAINER {
        S_vector Vec1;
        S_vector Vec2;
        S_vector Vec3;
        int unit;
    };
    typedef std::vector<GNSS_RESIDUAL_CONTAINER> GNSS_RESIDUAL_VECTOR;
	
    struct IMU_RESIDUAL_CONTAINER {
        S_vector Vec1;
        S_vector Vec2;
        S_vector Vec3;
        S_vector Vec4;
        S_vector Vec5;
        S_vector Vec6;
        int unit;
    };
    typedef std::vector<IMU_RESIDUAL_CONTAINER> IMU_RESIDUAL_VECTOR;

private:
	ros::NodeHandle nh;
	
	ros::Subscriber gnssGGA1_sub;
	ros::Subscriber gnssGGA2_sub;
	ros::Subscriber gnssGGA3_sub;
	ros::Subscriber IMU1_sub;
	ros::Subscriber IMU2_sub;
	ros::Subscriber IMU3_sub;
    ros::Subscriber gnss_outlierReport_sub;
    ros::Subscriber imu_outlierReport_sub;
    ros::Subscriber residual_pose_sub;
    ros::Subscriber residual_imu_sub;
    ros::Subscriber residual_generator_ready_sub;
    ros::Subscriber residual_generator_ready_IMU_sub;
    
    // Publishers
	ros::Publisher GNSS_posteriori_pub;
	ros::Publisher gnss_outlier_pub;
	ros::Publisher gnss_residual_pub1;
	ros::Publisher gnss_residual_pub2;
	ros::Publisher gnss_residual_pub3;
	
	ros::Publisher IMU_posteriori_pub;
	ros::Publisher imu_residual_pub1;
	ros::Publisher imu_residual_pub2;
	ros::Publisher imu_residual_pub3;
    ros::Publisher imu_outlier_pub;

	// Callback functions
	void gnssCB(const custom_msgs::gpsFIX::ConstPtr& msg, const int GNSS_Num);
	void imuCB(const custom_msgs::inertialSensor::ConstPtr& msg, const int IMU_Num);
    void GNSSoutlierCB(const custom_msgs::gnssOutlier &msg);
	void gnssResidualCB(const custom_msgs::residualPose::ConstPtr& msg);
    void IMUoutlierCB(const custom_msgs::imuOutlier &msg);
    void residualReadyCB(const std_msgs::UInt8 &msg);
	void IMUresidualReadyCB(const std_msgs::UInt8 &msg);
	void imuResidualCB(const custom_msgs::residualIMU::ConstPtr& msg);

	// Init functions
	void initPublishers();
	void initSubscribers();
	void initGNSSContainers();
	void initIMUContainers();
    
	int vote(double &reading1,double &reading2,double &reading3, double tolerance);
	std::vector<double> averageGNSS();
	std::vector<double> averageIMU();
	double averageVector(std::vector<double> const& input);
	double LowPassFilter(double input, double Output_last);
	template<typename T, typename A>
    void timestampChecker(std::vector<T,A> const& data_vec, int sensor_type);
	bool doubleEqual(double arg1, double arg2, double epsilon);
	std::vector<double> CUSUM_test(double residual, double u_H1,double u_H2, double sigma_null, double sigma_H1, double sigma_H2, std::vector<double> S_values);
	std::vector<std::vector<double>> VarianceGNSS(int samples);
	std::vector<std::vector<double>> VarianceIMU(int samples);
	bool sensor_reactivator(std::vector<SIG_INFO> unit_health, int unit);
	void gnssPub(std::vector<double> publish);
	void imuPub(std::vector<double> publish);
	void gnssFTC();
	void imuFTC();

	// GNSS Data Containers
	GNSS_DATA_VECTOR GNSS_Data;
    GNSS_RESIDUAL_VECTOR GNSS_Residual_Data;
    std::vector<double> gnss_treated{0.0,0.0,0.0};
    SIG_INFO_VECTOR GNNS_SignalHealth;

	// IMU Data Container
    IMU_DATA_VECTOR IMU_Data;
    IMU_RESIDUAL_VECTOR IMU_Residual_Data;
    std::vector<double> imu_treated{0.0,0.0,0.0,0.0,0.0,0.0};
    SIG_INFO_VECTOR IMU_SignalHealth;
    
    // GNSS FTCC variables
    double GNSS_max_var_horizontal = 0.000000075;
    double GNSS_max_var_vertical = 0.6;
    bool GNSS_voting_warning_printed = false;
    double beta_LP = 0.05; // LP filter constant
    double GNSS_lowpass_time = 5.0;
    double GNSS_lp_timelapsed = 0.0;
    double GNSS_time_last_lp = 0.0;
    std::vector<double> GNSS_LP_output_last{0,0,0};
    std::vector<int> GNSS_votes_last{0,0,0};
	double voting_threshold_horizontal = 0.005;
	double voting_threshold_vertical = 1;
    int operational_GNSS = NUM_GNSS_UNITS;
    bool GNSS_voted_out = false; 

    // IMU FTCC variables
    double IMU_max_var_angles = 2.5;
    double IMU_max_var_rates = 0.5;
    bool IMU_voting_warning_printed = false;
    double IMU_lowpass_time = 5.0;
    double IMU_lp_timelapsed = 0.0;
    double IMU_time_last_lp = 0.0;
    std::vector<double> IMU_LP_output_last{0,0,0};
    std::vector<int> IMU_votes_last{0,0,0,0,0,0};
	double voting_threshold_angles = 3;
	double voting_threshold_rates = 1;
    int operational_IMU = NUM_IMU_UNITS;
    bool IMU_voted_out = false; 
    
	// IMU CUSUM variables: 
	int IMU_residualsReady = 0;
	int IMU_threshold_high = 10;
	int IMU_threshold_low = -10;
	double u_H1_IMU = 0.2;
	double u_H2_IMU = -0.2;
	double sigma_null_IMU = 0.035;
	double sigma_H1_IMU = 0.055;
	double sigma_H2_IMU = 0.055;
	
	// GNSS CUSUM variables: 
	int GNSS_residualsReady = 0;
	int GNSS_threshold_high = 5;
	int GNSS_threshold_low = -5;
	double u_H1_GNSS_horizontal = 0.055;
	double u_H2_GNSS_horizontal = -0.055;
	double sigma_null_GNSS_horizontal = 0.00085;
	double sigma_H1_GNSS_horizontal = 0.03;
	double sigma_H2_GNSS_horizontal = 0.03;
	// Values for altitude has been shifted, as this measurement has higher uncertainty
	double u_H1_GNSS_vertical = 1.2;
	double u_H2_GNSS_vertical = -1.2;
	double sigma_null_GNSS_vertical = 0.625;
	double sigma_H1_GNSS_vertical = 0.2;
	double sigma_H2_GNSS_vertical = 0.2;
};
